/**
 * 响应式系统模块
 */

let activeFn = null
const targetMap = new WeakMap()

class Depend {
  constructor() {
    this.activeFns = new Set()
  }
  depend(){
    if(activeFn) this.activeFns.add(activeFn)
  }
  notify(){
    this.activeFns.forEach(fn => fn())
  }
}

function getDepend(target, key) {
  let depMap = targetMap.get(target)
  if(!depMap) {
    depMap = new Map()
    targetMap.set(target, depMap)
  }
  let dep = depMap.get(key)
  if(!dep) {
    dep = new Depend()
    depMap.set(key, dep)
  }
  return dep
}

function isObject(value) {
  const type = typeof value
  return (type === 'object' || type === 'function') && value !== null
}

/* 基于Proxy实现(Vue3) */
function reactive(obj, deep = true) {
  return new Proxy(obj, {
    get: function (target, key, receiver) {
      const dep = getDepend(target, key)
      dep.depend()
      const res = Reflect.get(target, key, receiver)
      /* deep: true -> 递归调用 */
      if(deep && isObject(res)) return reactive(res)
      return res
    },
    set: function (target, key, value, receiver) {
      Reflect.set(target, key, value, receiver)
      const dep = getDepend(target, key)
      dep.notify()
      return true
    }
  })
}

/* 基于Object.defineProperty实现(Vue2) */
// function reactive(obj) {
//   Object.keys(obj).forEach(key => {
//     let value = obj[key]
//     Object.defineProperty(obj, key, {
//       get: function () {
//         const dep = getDepend(obj, key)
//         dep.depend()
//         return value
//       },
//       set: function (newValue) {
//         value = newValue
//         const dep = getDepend(obj, key)
//         dep.notify()
//       }
//     })
//   })
//   return obj
// }

function watchEffect(fn) {
  activeFn = fn
  fn()
  activeFn = null
}

class Ref {
  constructor(initialValue) {
    this.value = initialValue
  }
  get value() {
    const dep = getDepend(this, "value")
    dep.depend()
    return this._value
  }
  set value(newValue) {
    this._value = newValue
    const dep = getDepend(this, "value")
    dep.notify()
  }
}

function ref(initialValue) {
  return new Ref(initialValue)
}

function computed(callback) {
  const computedRef = new Ref()
  watchEffect(() => {
    computedRef.value = callback()
  })
  return computedRef
}

export {
  reactive,
  watchEffect,
  ref,
  computed
}

/**
 * test demo
 */

/** test case 1 */
// const data = reactive({
//   name: "Michael",
//   age: 18,
//   friends: [{
//     name: "Sara",
//     age: 19
//   }]
// })

// watchEffect(() => console.log(data.name))
// watchEffect(() => console.log(data.age))
// watchEffect(() => console.log(data.friends[0].name))
// watchEffect(() => console.log(data.friends[0].age))

// data.name = "Zhiping Li"
// data.age++
// data.friends[0].name = "Rachael"
// data.friends[0].age += 2


/** test case 2 */
// const node = reactive({
//   leftChildren: 1,
//   // rightChildren: 0
// })
// console.log(node.leftChildren, node.rightChildren) // 1 undefined
// const children = computed(() => node.leftChildren + (parseInt(node.rightChildren) || 0))
// console.log(children.value) // 1
// node.leftChildren = 10
// console.log(children.value) // 10
// node.rightChildren = 2
// console.log(children.value) // 12


/** test case 3 */
// const numRef = ref(1)
// watchEffect(() => console.log(numRef.value))
// numRef.value++
